@@ -7,29 +7,39 @@
 # License:  Standard 3-clause BSD; see "license.txt" for full license terms
 #           and contributor agreement.
 
-import weakref
+"""
+    Implements high-level support for HDF5 file objects.
+"""
+
+from __future__ import absolute_import
+
 import sys
 import os
 
-from .base import HLObject, py3
+import six
+
+from .base import phil, with_phil
 from .group import Group
-from h5py import h5, h5f, h5p, h5i, h5fd, h5t, _objects
-from h5py import version
+from .. import h5, h5f, h5p, h5i, h5fd, _objects
+from .. import version
 
 mpi = h5.get_config().mpi
 hdf5_version = version.hdf5_version_tuple[0:3]
 
+swmr_support = False
+if hdf5_version >= h5.get_config().swmr_min_hdf5_version:
+    swmr_support = True
+
 if mpi:
     import mpi4py
 
 libver_dict = {'earliest': h5f.LIBVER_EARLIEST, 'latest': h5f.LIBVER_LATEST}
-libver_dict_r = dict((y, x) for x, y in libver_dict.iteritems())
+libver_dict_r = dict((y, x) for x, y in six.iteritems(libver_dict))
 
 
 def make_fapl(driver, libver, **kwds):
     """ Set up a file access property list """
     plist = h5p.create(h5p.FILE_ACCESS)
-    plist.set_fclose_degree(h5f.CLOSE_STRONG)
 
     if libver is not None:
         if libver in libver_dict:
@@ -42,15 +52,15 @@ def make_fapl(driver, libver, **kwds):
     if driver is None or (driver == 'windows' and sys.platform == 'win32'):
         return plist
 
-    if(driver == 'sec2'):
+    if driver == 'sec2':
         plist.set_fapl_sec2(**kwds)
-    elif(driver == 'stdio'):
+    elif driver == 'stdio':
         plist.set_fapl_stdio(**kwds)
-    elif(driver == 'core'):
+    elif driver == 'core':
         plist.set_fapl_core(**kwds)
-    elif(driver == 'family'):
+    elif driver == 'family':
         plist.set_fapl_family(memb_fapl=plist.copy(), **kwds)
-    elif(driver == 'mpio'):
+    elif driver == 'mpio':
         kwds.setdefault('info', mpi4py.MPI.Info())
         plist.set_fapl_mpio(**kwds)
     else:
@@ -59,7 +69,7 @@ def make_fapl(driver, libver, **kwds):
     return plist
 
 
-def make_fid(name, mode, userblock_size, fapl, fcpl=None):
+def make_fid(name, mode, userblock_size, fapl, fcpl=None, swmr=False):
     """ Get a new FileID by opening or creating a file.
     Also validates mode argument."""
 
@@ -76,10 +86,13 @@ def make_fid(name, mode, userblock_size, fapl, fcpl=None):
         fcpl.set_userblock(userblock_size)
 
     if mode == 'r':
-        fid = h5f.open(name, h5f.ACC_RDONLY, fapl=fapl)
+        flags = h5f.ACC_RDONLY
+        if swmr and swmr_support:
+            flags |= h5f.ACC_SWMR_READ
+        fid = h5f.open(name, flags, fapl=fapl)
     elif mode == 'r+':
         fid = h5f.open(name, h5f.ACC_RDWR, fapl=fapl)
-    elif mode == 'w-':
+    elif mode in ['w-', 'x']:
         fid = h5f.create(name, h5f.ACC_EXCL, fapl=fapl, fcpl=fcpl)
     elif mode == 'w':
         fid = h5f.create(name, h5f.ACC_TRUNC, fapl=fapl, fcpl=fcpl)
@@ -103,7 +116,7 @@ def make_fid(name, mode, userblock_size, fapl, fcpl=None):
             except IOError:
                 fid = h5f.create(name, h5f.ACC_EXCL, fapl=fapl, fcpl=fcpl)
     else:
-        raise ValueError("Invalid mode; must be one of r, r+, w, w-, a")
+        raise ValueError("Invalid mode; must be one of r, r+, w, w-, x, a")
 
     try:
         if userblock_size is not None:
@@ -124,14 +137,16 @@ class File(Group):
     """
 
     @property
+    @with_phil
     def attrs(self):
         """ Attributes attached to this object """
         # hdf5 complains that a file identifier is an invalid location for an
         # attribute. Instead of self, pass the root group to AttributeManager:
-        import attrs
+        from . import attrs
         return attrs.AttributeManager(self['/'])
 
     @property
+    @with_phil
     def filename(self):
         """File name on disk"""
         name = h5f.get_name(self.fid)
@@ -141,6 +156,7 @@ class File(Group):
             return name
 
     @property
+    @with_phil
     def driver(self):
         """Low-level HDF5 file driver used to open file"""
         drivers = {h5fd.SEC2: 'sec2', h5fd.STDIO: 'stdio',
@@ -150,23 +166,27 @@ class File(Group):
         return drivers.get(self.fid.get_access_plist().get_driver(), 'unknown')
 
     @property
+    @with_phil
     def mode(self):
         """ Python mode used to open file """
         return {h5f.ACC_RDONLY: 'r',
                 h5f.ACC_RDWR: 'r+'}.get(self.fid.get_intent())
 
     @property
+    @with_phil
     def fid(self):
         """File ID (backwards compatibility) """
         return self.id
 
     @property
+    @with_phil
     def libver(self):
         """File format version bounds (2-tuple: low, high)"""
         bounds = self.id.get_access_plist().get_libver_bounds()
         return tuple(libver_dict_r[x] for x in bounds)
 
     @property
+    @with_phil
     def userblock_size(self):
         """ User block size (in bytes) """
         fcpl = self.fid.get_create_plist()
@@ -176,18 +196,36 @@ class File(Group):
     if mpi and hdf5_version >= (1, 8, 9):
 
         @property
+        @with_phil
         def atomic(self):
             """ Set/get MPI-IO atomic mode 
             """
             return self.id.get_mpi_atomicity()
 
         @atomic.setter
+        @with_phil
         def atomic(self, value):
+            # pylint: disable=missing-docstring
             self.id.set_mpi_atomicity(value)
+            
+    if swmr_support:
+        @property
+        def swmr_mode(self):
+            """ Controls single-writer multiple-reader mode """
+            return self._swmr_mode
+            
+        @swmr_mode.setter
+        @with_phil
+        def swmr_mode(self, value):
+            # pylint: disable=missing-docstring
+            if value:
+                self.id.start_swmr_write()
+                self._swmr_mode = True
+            else:
+                raise ValueError("It is not possible to forcibly switch SWMR mode off.")
 
-
-    def __init__(self, name, mode=None, driver=None, 
-                 libver=None, userblock_size=None, **kwds):
+    def __init__(self, name, mode=None, driver=None,
+                 libver=None, userblock_size=None, swmr=False, **kwds):
         """Create a new file object.
 
         See the h5py user guide for a detailed explanation of the options.
@@ -195,6 +233,12 @@ class File(Group):
         name
             Name of the file on disk.  Note: for files created with the 'core'
             driver, HDF5 still requires this be non-empty.
+        mode
+            r        Readonly, file must exist
+            r+       Read/write, file must exist
+            w        Create file, truncate if exists
+            w- or x  Create file, fail if exists
+            a        Read/write if exists, create otherwise (default)
         driver
             Name of the driver to use.  Legal values are None (default,
             recommended), 'core', 'sec2', 'stdio', 'mpio'.
@@ -203,57 +247,89 @@ class File(Group):
             and 'latest' are defined.
         userblock
             Desired size of user block.  Only allowed when creating a new
-            file (mode w or w-).
+            file (mode w, w- or x).
+        swmr
+            Open the file in SWMR read mode. Only used when mode = 'r'.
         Additional keywords
             Passed on to the selected file driver.
         """
-        if isinstance(name, _objects.ObjectID):
-            fid = h5i.get_file_id(name)
-        else:
-            try:
-                # If the byte string doesn't match the default
-                # encoding, just pass it on as-is.  Note Unicode
-                # objects can always be encoded.
-                name = name.encode(sys.getfilesystemencoding())
-            except (UnicodeError, LookupError):
-                pass
+        if swmr and not swmr_support:
+            raise ValueError("The SWMR feature is not available in this version of the HDF5 library")
+        
+        with phil:
+            if isinstance(name, _objects.ObjectID):
+                fid = h5i.get_file_id(name)
+            else:
+                try:
+                    # If the byte string doesn't match the default
+                    # encoding, just pass it on as-is.  Note Unicode
+                    # objects can always be encoded.
+                    name = name.encode(sys.getfilesystemencoding())
+                except (UnicodeError, LookupError):
+                    pass
 
-            fapl = make_fapl(driver, libver, **kwds)
-            fid = make_fid(name, mode, userblock_size, fapl)
-
-        Group.__init__(self, fid)
+                fapl = make_fapl(driver, libver, **kwds)
+                fid = make_fid(name, mode, userblock_size, fapl, swmr=swmr)
+            
+                if swmr_support:
+                    self._swmr_mode = False
+                    if swmr and mode == 'r':
+                        self._swmr_mode = True                    
+                    
+            Group.__init__(self, fid)
 
     def close(self):
         """ Close the file.  All open objects become invalid """
-        # TODO: find a way to square this with having issue 140
-        # Not clearing shared state introduces a tiny memory leak, but
-        # it goes like the number of files opened in a session.
-        self.id.close()
+        with phil:
+            # We have to explicitly murder all open objects related to the file
+            
+            # Close file-resident objects first, then the files.
+            # Otherwise we get errors in MPI mode.
+            id_list = h5f.get_obj_ids(self.id, ~h5f.OBJ_FILE)
+            file_list = h5f.get_obj_ids(self.id, h5f.OBJ_FILE)
+            
+            id_list = [x for x in id_list if h5i.get_file_id(x).id == self.id.id]
+            file_list = [x for x in file_list if h5i.get_file_id(x).id == self.id.id]
+            
+            for id_ in id_list:
+                while id_.valid:
+                    h5i.dec_ref(id_)
+                    
+            for id_ in file_list:
+                while id_.valid:
+                    h5i.dec_ref(id_)
+                    
+            self.id.close()
+            _objects.nonlocal_close()
 
     def flush(self):
         """ Tell the HDF5 library to flush its buffers.
         """
-        h5f.flush(self.fid)
+        with phil:
+            h5f.flush(self.fid)
 
+    @with_phil
     def __enter__(self):
         return self
 
+    @with_phil
     def __exit__(self, *args):
         if self.id:
             self.close()
 
+    @with_phil
     def __repr__(self):
         if not self.id:
-            r = u'<Closed HDF5 file>'
+            r = six.u('<Closed HDF5 file>')
         else:
             # Filename has to be forced to Unicode if it comes back bytes
             # Mode is always a "native" string
             filename = self.filename
             if isinstance(filename, bytes):  # Can't decode fname
                 filename = filename.decode('utf8', 'replace')
-            r = u'<HDF5 file "%s" (mode %s)>' % (os.path.basename(filename),
+            r = six.u('<HDF5 file "%s" (mode %s)>') % (os.path.basename(filename),
                                                  self.mode)
 
-        if py3:
+        if six.PY3:
             return r
         return r.encode('utf8')
